#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["SingletonBase", "SingletonMeta", "key_singleton", "singleton"]

import functools
from functools import wraps
from typing import Any


class SingletonBase:
  """
    A base class for creating singleton class where all subtypes(Deriavations) should also return the first
    and only
    instantiation of a particular singleton base type, if this property is not wanted consider using the
    SingletonMeta class instead."""

  instance = None

  def __new__(cls, *args, **kwargs):
    if cls.instance is None:
      cls.instance = super().__new__(cls, *args, **kwargs)

    return cls.instance


class SingletonMeta(type):
  """
    Conversely the SingletonBase, this base meta class is used for creating singleton class where all
    subtypes(
    Deriavations) should only
    return
    singleton instantiations of a particular singleton type independantly of subtyping and super-types,
    if this property is not
    wanted
    consider using
    the
    SingletonBase class instead."""

  def __init__(cls, name, bases, namespace):
    super().__init__(name, bases, namespace)
    cls.instance = None

  def __call__(cls, *args, **kwargs):
    if cls.instance is None:
      cls.instance = super().__call__(*args, **kwargs)

    return cls.instance


def singleton(cls):
  """ Use class as singleton. """

  cls.__new_original__ = cls.__new__

  @functools.wraps(cls.__new__)
  def singleton_new(cls_, *args, **kw):
    """

        :param cls_:
        :param args:
        :param kw:
        :return:
        """
    it = cls_.__dict__.get("__it__")
    if it is not None:
      return it

    cls_.__it__ = it = cls_.__new_original__(cls_, *args, **kw)
    it.__init_original__(*args, **kw)
    return it

  cls.__new__ = singleton_new
  cls.__init_original__ = cls.__init__
  cls.__init__ = object.__init__

  return cls


def key_singleton(cache_key:Any)->callable:
  """
  TODO: finish
    """

  def inner_fn(fn:callable)->Any:
    """
        """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
      """
            """
      instance = getattr(self, cache_key)
      if instance is not None:
        return instance

      instance = fn(self, *args, **kwargs)
      setattr(self, cache_key, instance)
      return instance

    return wrapper

  return inner_fn


if __name__ == "__main__":

  class SingletonBaseClass(SingletonBase):
    pass


  class S1(SingletonBaseClass):
    pass


  class SingletonBaseMeta(metaclass=SingletonMeta):
    pass


  class S2(SingletonBaseMeta):
    pass


  # expected
  print(SingletonBaseClass())  # same
  print(SingletonBaseClass())  # same
  print(S1())  # same

  # expected
  print(SingletonBaseMeta())  # same
  print(SingletonBaseMeta())  # same
  print(S2())  # different



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
from typing import Any, Mapping, MutableMapping

__author__ = "Christian Heider Nielsen"
__doc__ = '''

Generalised wrapper for delayed construction of class objects. Encapsulates kwargs and callable constructor with the option of modifying construction arguments before construction is finally performed.

'''

__all__ = ["GeneralisedDelayedKwargConstruction", "GDKC"]


class GeneralisedDelayedKwargConstruction(object):
  """
  A generalised class for setting up kwargs for later construction of an instance of an object
  [constructor, args, kwargs]
  """

  def __init__(self, constructor: callable, *args, **kwargs):
    """
    [constructor, args, kwargs]
    :param constructor:
    :param args:
    :param kwargs:"""
    self.constructor: callable = constructor
    assert (
        len(args) < 2
    ), f"Does not support multiple args, only a single mapping type"
    if len(args) == 1:
      assert isinstance(
          args[0], Mapping
          ), f"Arg[0] type is not a mapping type, was {type(args[0])} which is not supported"
    assert not (
        len(kwargs) > 0 and len(args) > 0
    ), f"Does not support both args and kwargs, both supplied args, {args} and {kwargs}"

    if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], Mapping):
      self.kwargs: MutableMapping = args[0]
    elif len(kwargs) == 1 and next(iter(kwargs.keys())) == "kwargs":
      self.kwargs = kwargs["kwargs"]
    else:
      self.kwargs: MutableMapping = kwargs

  def __enter__(self):
    self.instance = self.__call__()
    return self.instance.__enter__()

  def __exit__(self, exc_type, exc_val, exc_tb):
    return self.instance.__exit__(exc_type, exc_val, exc_tb)

  def __call__(self, *args, **kwargs) -> Any:
    """

    :param args:
    :param kwargs:
    :return:"""

    war = "".join(
        [
            f"Overwriting {k} with the value {v} in construction of {self.constructor}"
            for (k, v), b in zip(kwargs.items(), self.kwargs.keys())
            if k == b
            ]
        )
    if war != "":
      logging.warning(war)

    self.kwargs.update(kwargs)
    try:
      return self.constructor(*args, **self.kwargs)
    except TypeError as e:
      e.args += (f"in construction of {self.constructor}",)
      raise e


GDKC = GeneralisedDelayedKwargConstruction

if __name__ == "__main__":

  class UnreachableError(Exception):
    pass


  class A:
    """
    """

    def __init__(self, *args, **kwargs):
      pass


  def stest_not_both():
    GeneralisedDelayedKwargConstruction(A, [1], a=2)


  def stest_kw():
    GeneralisedDelayedKwargConstruction(A, a=2)


  def stest_mapping():
    GeneralisedDelayedKwargConstruction(A, {"a":2})


  def stest_mapping_and_args_fail():
    GeneralisedDelayedKwargConstruction(A, {"a":2}, 1)


  def stest_mapping_and_args_fail_inv():
    GeneralisedDelayedKwargConstruction(A, 1, {"a":2})


  stest_kw()
  stest_mapping()

  try:
    stest_mapping()  # Successful
    raise UnreachableError
  except UnreachableError:  # Is expected
    pass
  try:
    stest_mapping_and_args_fail()
    raise UnreachableError
  except AssertionError:
    pass
  try:
    stest_mapping_and_args_fail_inv()
    raise UnreachableError
  except AssertionError:
    pass
  try:
    stest_not_both()
    raise UnreachableError
  except AssertionError:
    pass
